"""
Mailer backend that writes email messages to a buffer (list) instead of sending
them. Useful for testing.
"""

import threading

from manhattan.mail import EmailTemplate
from manhattan.mail.backends.base import BaseMailer

__all__ = ['Mailer']


class Mailer(BaseMailer):
    """
    Mailer that writes emails to a buffer (list).
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._lock = threading.RLock()

        if not hasattr(self.__class__, 'outbox'):
            self.__class__.outbox = []

    def clear(self):
        """Clear the outbox"""
        if hasattr(self.__class__, 'outbox'):
            self.__class__.outbox = []

    def send(self, messages):
        """Write all email messages to the buffer"""
        if isinstance(messages, EmailTemplate):
            messages = messages.merge()

        message_count = 0

        with self._lock:
            for message in messages:
                self.__class__.outbox.append(message.message)
                message_count += 1

        return message_count



from email import charset, encoders
from email.header import Header
from email.headerregistry import Address
from email.mime.application import MIMEApplication
from email.mime.audio import MIMEAudio
from email.mime.base import MIMEBase
from email.mime.image import MIMEImage
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.utils import formatdate, make_msgid
import mimetypes
import os
import re
import socket

import html2text
from jinja2 import Environment, FileSystemLoader
from premailer import Premailer

__all__ = [
    'EmailAttachment',
    'EmailContact',
    'EmailMessage',
    'EmailTemplate'
    ]

# Don't BASE64-encode UTF-8 messages so that we avoid unwanted attention from
# some spam filters.
utf8_charset = charset.Charset('utf-8')
utf8_charset.body_encoding = None
utf8_charset_qp = charset.Charset('utf-8')
utf8_charset_qp.body_encoding = charset.QP

# Patch mimetypes lib to recognize CSV files
mimetypes.add_type('text/csv', '.csv')


class EmailAttachment:
    """
    Information about an email attachment.
    """

    BASETYPE_MAP = {
        'application': MIMEApplication,
        'audio': MIMEAudio,
        'image': MIMEImage,
        'text': MIMEText
    }
    DEFAULT_MIMETYPE = 'application/octet-stream'

    def __init__(self, filename, content, mimetype=None):
        self.filename = filename
        self.content = content

        # Determine the mimetype
        if not mimetype:
            mimetype, _ = mimetypes.guess_type(filename)
            if not mimetype:
                mimetype = self.DEFAULT_MIMETYPE

        self.mimetype = mimetype

        # If the base type is text check that the content isn't binary
        basetype, subtype = self.mimetype.split('/', 1)

        if basetype == 'text':
            if isinstance(self.content, bytes):
                try:
                    self.content = self.content.decode()
                except UnicodeDecodeError:
                    self.mimetype = self.DEFAULT_MIMETYPE

    @property
    def attachment(self):
        """Return an email attachment"""
        attachment = None

        # Build the MIME object
        basetype, subtype = self.mimetype.split('/', 1)
        if basetype in self.BASETYPE_MAP:
            attachment = self.BASETYPE_MAP[basetype](self.content, subtype)
        else:
            attachment = MIMEBase(basetype, subtype)
            attachment.set_payload(self.content)
            encoders.encode_base64(attachment)

        # Set the filename
        attachment.add_header(
            'Content-Disposition',
            'attachment',
            filename=self.filename
            )

        return attachment

    @classmethod
    def from_file(cls, file_path):
        """Create an attachment from a file"""
        filename = None
        content = None

        if isinstance(file_path, str):
            filename = file_path
            with open(file_path, 'rb') as f:
                content = f.read()

        else:
            filename = file_path.name
            file_path.seek(0)
            content = file_path.read()

        return cls(os.path.basename(filename), content)


class EmailContact:
    """
    Information about a email recipient or sender.
    """

    def __init__(self, addr, name='', encoding='utf-8'):
        self.addr = addr
        self.name = name
        self.encoding = encoding

    def __eq__(self, other):
        return self.addr == other.addr

    def __hash__(self):
        return hash(self.addr)

    def __iter__(self):
        for v in (self.addr, self.name, self.encoding):
            yield v

    def __repr__(self):
        return str(self)

    def __str__(self):
        """Return an email contact"""
        username, domain = self.addr.split('@', 1)
        try:
            username.encode('ascii')
        except UnicodeEncodeError:
            username = Header(username, self.encoding).encode()
        domain = domain.encode('idna').decode('ascii')
        addr = Address(self.name, username=username, domain=domain)
        return str(addr)

    @classmethod
    def normalize(cls, addrs, default_encoding='utf-8'):
        """
        Normalize one or more addresses of the form string, list/tuple or
        `EmailContact`) to `EmailContact`s.
        """
        normalized = []
        for addr in addrs:
            if isinstance(addr, cls):
                addr = addr
            elif isinstance(addr, (list, tuple)):
                addr = cls(*addr)
            else:
                addr = cls(addr)
            normalized.append(addr)
        return normalized


class EmailMessage:
    """
    Information about an email message.
    """

    LINE_LENGTH_LIMIT = 998

    def __init__(
            self,
            to,
            sender,
            subject,
            body_text=None,
            body_html=None,
            cc=None,
            bcc=None,
            attachments=None,
            headers=None,
            encoding='utf-8'
            ):

        # A list of recipients for the email
        self.to = EmailContact.normalize(to)

        # A list of contacts who will be CC'd on this email
        self.cc = EmailContact.normalize(cc or [])

        # A list of contacts who will be bCC'd on this email
        self.bcc = EmailContact.normalize(bcc or [])

        # The contact the email is being sent from
        self.sender = EmailContact.normalize([sender])[0]

        # The subject line for the email
        self.subject = subject

        # The body of the email in plain text format
        self.body_text = body_text

        # The body of the email in HTML format
        self.body_html = body_html

        # If a plain text body hasn't been provided then create one
        if self.body_text is None and self.body_html:
            text_maker = html2text.HTML2Text()
            self.body_text = text_maker.handle(self.body_html)

        # A list of attachments to be attached to the message
        self.attachments = attachments or []

        # A list of additional headers for the message
        self.headers = headers or {}

        # The charet encoding for the body of the email
        self.encoding = encoding

    def attach(self, attachment):
        """
        Add an attachment to the email message, the attachment can be an
        `EmailAttachment` instance, a file point or file path.
        """
        if isinstance(attachment, EmailAttachment):
            self.attachments.append(attachment)
        else:
            attachment = EmailAttachment.from_file(attachment)
            self.attachments.append(attachment)
        return attachment

    @property
    def message(self):
        """Return an email message"""

        # Build the text message
        msg = text_msg = self._safe_mime_text(
            self.body_text,
            'plain',
            self.encoding
            )

        # Build the HTML message
        if self.body_html:
            html_msg = self._safe_mime_text(
                self.body_html,
                'html',
                self.encoding
                )

        # Handle attachments
        if len(self.attachments):
            msg = MIMEMultipart(_subtype='mixed', encoding=self.encoding)

            # If a HTML version of the email is provided then we need to create
            # a text and HTML message to attach to the primary message along
            # with the attachments...
            if self.body_html:
                text_html_msg = MIMEMultipart(
                    _subtype='alternative',
                    encoding=self.encoding
                    )
                text_html_msg.attach(text_msg)
                text_html_msg.attach(html_msg)
                msg.attach(text_html_msg)

            # ...else we simple attact the text.
            else:
                msg.attach(text_msg)

            # Add any attachments
            for attachment in self.attachments:
                msg.attach(attachment.attachment)

        elif self.body_html:
            # If a HTML version is provided then we need to create a text and
            # HTML message as the primary message.
            msg = MIMEMultipart(_subtype='alternative', encoding=self.encoding)
            msg.attach(text_msg)
            msg.attach(html_msg)

        # Add common headers
        msg['Subject'] = self.subject
        msg['From'] = str(self.sender)
        msg['To'] = ', '.join([str(r) for r in self.to])
        if self.cc:
            msg['Cc'] = ', '.join([str(r) for r in self.cc])
        msg['Date'] = formatdate()
        msg['Message-ID'] = make_msgid(socket.getfqdn())

        # Add extra headers
        header_names = [k.lower() for k in self.headers]
        for name, value in self.headers.items():
            if name.lower() in ('from', 'to'):
                continue
            msg[name] = value

        return msg

    @property
    def recipients(self):
        """Return a list of all recipients for the email message"""
        return self.to + self.cc + self.bcc

    def _safe_mime_text(self, payload, subtype, charset):
        """Return a MIMEText message that is safe to send"""

        # Check for utf-8 encoding
        if charset == 'utf-8':
            charset = utf8_charset

            # Check for long lines
            has_long_lines = any(
                len(l.encode()) > self.LINE_LENGTH_LIMIT
                for l in payload.splitlines()
                )
            if has_long_lines:
                charset = utf8_charset_qp

        # Set the payload
        msg = MIMEText('', subtype)
        msg.set_payload(payload, charset=charset)

        return msg

class EmailTemplate:
    """
    A class for generating one or more `EmailMessage` by mail merging a template
    with a list of recipients.
    """

    def __init__(
            self,
            to,
            sender,
            subject,
            template_path,
            template_map=None,
            css='',
            recipient_vars=None,
            global_vars=None,
            cc=None,
            bcc=None,
            attachments=None,
            headers=None,
            base_url=None,
            format='html',
            encoding='utf-8'
            ):

        # A list of recipients for the email
        self.to = EmailContact.normalize(to)

        # A list of contacts who will be CC'd on this email
        self.cc = EmailContact.normalize(cc or [])

        # A list of contacts who will be bCC'd on this email
        self.bcc = EmailContact.normalize(bcc or [])

        # The contact the email is being sent from
        self.sender = EmailContact.normalize([sender])[0]

        # The subject for the email. The subject string is compiled as a
        # template and can therefore be personalized using recipient and global
        # vars.
        self.subject = subject

        # The path to the template
        self.template_path = template_path

        # Template maps allow the template (and the templates it depends on) to
        # be provided as strings (instead of relying on the file system).
        self.template_map = template_map

        # A string of CSS that that will be inlined into each email messages
        # HTML body (only relevant when the format is HTML).
        self.css = css

        # A dictionary of recipient data where each recipients data is stored
        # against the recipients email address, e.g:
        #
        #     {recipient email: {recipient data}}.
        #
        # When an email message is compiled for a recipient, the message
        # recipient's data is provided to the email template so that the message
        # and subject can be personalized.
        #
        self.recipient_vars = recipient_vars or {}

        # A dictionary of data provided to when compiling the template and
        # subject for all messages.
        self.global_vars = global_vars or {}

        # A list of attachments to attach to any email message generated by the
        # template.
        self.attachments = attachments or []

        # A list of headers to include in any email message generated by the
        # template.
        self.headers = headers or {}

        # The base URL to prefix to all relative URLs within the email
        self.base_url = base_url

        # The format the template will output email messages as (either 'html'
        # or 'text'). The default is 'html', however, at text alternative is
        # automatically generated for HTML email messages.
        self.format = format

        # The encoding used when the template outputs an email message
        self.encoding = encoding

    def attach(self, attachment):
        """
        Add an attachment to the email template, the attachment can be an
        `EmailAttachment` instance, a file point or file path.
        """
        if isinstance(attachment, EmailAttachment):
            self.attachments.append(attachment)
        else:
            attachment = EmailAttachment.from_file(attachment)
            self.attachments.append(attachment)
        return attachment

    def get_jinja_env(self):
        """
        Return an environment to render the template with (this method should
        be overridden if you want to provide a custom environment).
        """
        return Environment(
            loader=FileSystemLoader(''),
            autoescape=True
            )

    def inline_css(self, body, css):
        """
        Inline CSS into a body of HTML (this method should be overridden if you
        want to implement a custom CSS inliner).
        """
        premailer = Premailer(
            body,
            base_url=self.base_url,
            css_text=css,
            keep_style_tags=True,
            strip_important=False,
            include_star_selectors=True,
            disable_validation=True
            )
        return premailer.transform()

    def merge(self):
        """Generate a list of email messages from the template"""
        messages = []

        # Build the templates
        env = self.get_jinja_env()

        # Subject
        subject_template = env.from_string(self.subject)

        # Body

        # If a template map is provided use it to build the template
        # environment (render from memory).
        body_template = None
        if self.template_map:
            # Compile and store each template in the mapping accept the one
            # specified by the template path.
            templates = {}
            for name, template in self.template_map.items():
                if name == self.template_path:
                    continue
                templates[name] = env.from_string(template, templates)

            # Compile the email template
            body_template = env.from_string(
                self.template_map[self.template_path],
                templates
                )

        # ...else render the template from via the file system
        else:
            body_template = env.get_template(self.template_path)

        # Strip comments from the CSS before inlining it
        css = re.sub(r'(?s)/\*.*?\*/', '', self.css).strip()

        for recipient in self.to:

            # Build the template args
            template_args = {}
            template_args.update(self.global_vars)
            template_args.update(self.recipient_vars.get(recipient.addr, {}))

            # Render the subject
            subject = subject_template.render(**template_args)

            # Render the body
            body = body_template.render(**template_args)

            # Inline the CSS for the message
            if self.css and self.format == 'html':
                body = self.inline_css(body, css)

            # Create the message and add it to the list of mail merged messages
            messages.append(EmailMessage(
                to=[recipient],
                sender=self.sender,
                subject=subject,
                body_text=body if self.format == 'text' else None,
                body_html=body if self.format == 'html' else None,
                cc=self.cc,
                bcc=self.bcc,
                attachments=self.attachments,
                headers=self.headers,
                encoding=self.encoding
                ))

        return messages



import pytest

from manhattan.mail.backends import oakridge
try:
    from manhattan.mail.tests.oakridge_settings import settings
except:
    settings = {}

# Fixtures

@pytest.fixture(scope='function')
def mailer():
    mailer = oakridge.Mailer(**settings)
    return mailer


# Tests

def test_init():
    """Initialize an email contact"""
    mailer = oakridge.Mailer(**settings)
    assert isinstance(mailer, oakridge.Mailer)

def test_send_email_template(mailer, email_template):
    """Write an email template to memory (as opposed to actually sending them)"""
    response = mailer.send(email_template)

    assert response.status_code == 200


"""
Mailer backend that sends email via oak-ridge.

oak-ridge is an open-source transactional mailer application developed by
Getme with a simple API for sending tran
"""

from contextlib import contextmanager
import io
import base64
import json
import sys
import requests
import threading

from manhattan.mail import EmailTemplate
from manhattan.mail.backends.base import BaseMailer

__all__ = ['Mailer']


class Mailer(BaseMailer):
    """
    Mailer that sends emails via oak-ridge.
    """

    def __init__(self,
        api_key='',
        api_endpoint='https://oak-ridge2.getme.co.uk',
        api_version='1.0',
        *args,
        **kwargs
        ):

        self.api_key = api_key
        self.api_endpoint = api_endpoint
        self.api_version = api_version
        super().__init__(*args, **kwargs)

        # A table for capturing Ids during a send
        self._capture_ids = None

    @contextmanager
    def capture_ids(self):
        """
        Capture a map of Ids generated when sending emails via oakridge.

        Note: Oakridge returns a single Id per send (even when there are
        multiple recipients), however, we combine this Id with each recipient
        (email address) to provide a unique Id for each email sent.
        """

        try:
            self._capture_ids = {}
            yield self._capture_ids

        finally:
            self._capture_ids = None

    def send(self, template):
        """Send the message template using oak-ridge"""

        # Validate the template is configured correctly for oak-ridge
        assert isinstance(template, EmailTemplate), \
                'oak-ridge supports sending `EmailTemplate` instances only'

        assert template.template_map, \
                '`template_map` must be used when sending through oak-ridge'

        # Build the URL to call
        url = '{endpoint}/{version}/send-email'.format(
            endpoint=self.api_endpoint,
            version=self.api_version
        )

        # Build the payload
        payload = {
            'api_key': self.api_key,
            'base_url': template.base_url,
            'bcc': [(r.addr, r.name) for r in template.bcc],
            'cc': [(r.addr, r.name) for r in template.cc],
            'css': template.css,
            'global_vars': template.global_vars,
            'recipient_vars': template.recipient_vars,
            'sender': (template.sender.addr, template.sender.name),
            'subject': template.subject,
            'template_path': template.template_path,
            'template_map': template.template_map,
            'to': [(r.addr, r.name) for r in template.to]
        }

        # Check for attachments
        if len(template.attachments) > 0:
            payload['attachments'] = []
            for attachment in template.attachments:

                # If the attachment content is not already in bytes then
                # covert it to bytes
                if isinstance(attachment.content, bytes):
                    content = attachment.content
                else:
                    content = io.BytesIO(
                        attachment.content.encode('utf-8')
                    ).getbuffer()

                payload['attachments'].append(
                    {
                        'filename': attachment.filename,
                        'file': base64.b64encode(
                            content
                        ).decode('utf8')
                    }
                )

        # Call the API
        response = requests.post(url, data={'payload': json.dumps(payload)})

        # Check for successful send to generate Ids
        try:
            response_json = response.json()

            if response_json['status'] == 'success':

                # Generate a table of recipients and oakridge Ids for the send
                id = response_json['payload']['send_request']

                if self._capture_ids is not None:
                    self._capture_ids.update({
                        r.addr: f'{id}:{r.addr.lower()}'
                        for r in template.to
                    })

        except ValueError:
            pass

        return response



import pytest

from manhattan.mail.backends import memory


# Fixtures

@pytest.fixture(scope='function')
def mailer():
    mailer = memory.Mailer()
    mailer.clear()
    return mailer


# Tests

def test_init():
    """Initialize an email contact"""
    mailer = memory.Mailer()
    assert isinstance(mailer, memory.Mailer)

def test_clear(mailer, email_message):
    """Clear the outbox"""
    assert len(mailer.outbox) == 0

    mailer.send([email_message])
    assert len(mailer.outbox) == 1

    mailer.clear()
    assert len(mailer.outbox) == 0

def test_send_email_message(mailer, email_message):
    """Write an email message to memory (as opposed to actually sending them)"""
    mailer.send([email_message])

    assert len(mailer.outbox) == 1
    mailer.outbox[0] == email_message

def test_send_email_template(mailer, email_template):
    """Write an email template to memory (as opposed to actually sending them)"""
    email_messages = email_template.merge()

    mailer.send(email_template)

    assert len(mailer.outbox) == 2
    mailer.outbox[0] == email_messages[0]
    mailer.outbox[1] == email_messages[1]



"""
Classes for implementing a backend for asset management.
"""

import json

__all__ = ['BaseMailer']


# Classes

class BaseMailer:
    """
    Email is sent out by via a mailer class which must provide the API defined
    by the `BaseMailer` class.
    """

    def __init__(self, fail_silently=False, **kwargs):
        # Flag indicating if exceptions raised while sending email messages
        # should be ignored.
        self.fail_silently = fail_silently

    def __enter__(self):
        try:
            self.open()
        except Exception:
            self.close()
            raise
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def open(self):
        """Open a connection to the email server/service"""

    def close(self):
        """Close connection to the email server/service"""

    def send(self, messages):
        """
        Send a list of email messages. The `email_messages` argument can either
        be defined as a list of `EmailMessage` instances or an `EmailTemplate`
        instance.
        """
        raise NotImplementedError('Must be overridden by inheriting classes')



"""
Mailer backend that writes email messages to the console instead of sending
them.
"""

import os
import pathlib
import sys
import threading

from manhattan.mail import EmailTemplate
from manhattan.mail.backends.base import BaseMailer

__all__ = ['Mailer']


class Mailer(BaseMailer):
    """
    Mailer that writes emails messages to file.

    This backend is provided to help developers review emails before
    production, each email sent overrides the contents of a single file.
    """

    def __init__(self, file_path=None, *args, **kwargs):
        self.file_path = file_path
        self._lock = threading.RLock()

        assert self.file_path, 'A `file_path` must be specified'

        super().__init__(*args, **kwargs)

    def send(self, messages):
        """Write each email messages to a file given by the file path"""
        if isinstance(messages, EmailTemplate):
            messages = messages.merge()

        # Split the file path into a directory path and filename
        dir_path, filename = os.path.split(self.file_path)

        # Make sure the save directory exists
        pathlib.Path(dir_path).mkdir(parents=True, exist_ok=True)

        message_count = 0

        with self._lock:
            try:
                for message in messages:
                    with open(self.file_path, 'w') as f:
                        self._write_message(f, message)

                    message_count += 1
            except Exception:
                if not self.fail_silently:
                    raise

        return message_count

    def _write_message(self, file, message):
        """Write a message to the stream"""
        msg_data = message.message.as_bytes()
        msg_data = msg_data.decode(message.encoding)
        file.write('%s\n' % msg_data)
        file.write('-' * 79)
        file.write('\n')



from email.mime.text import MIMEText
import os

from manhattan.mail import EmailAttachment


def test_init():
    """Initialize an email attachment"""
    attachment = EmailAttachment('hello-world.txt', 'Hello world')
    assert isinstance(attachment, EmailAttachment)

def test_attachment():
    """Return an attachment message"""
    attachment = EmailAttachment('hello-world.txt', 'Hello world').attachment

    assert isinstance(attachment, MIMEText)
    assert attachment.get_payload() == 'Hello world'
    assert attachment['Content-Disposition'] == \
            'attachment; filename="hello-world.txt"'

def test_from_file():
    """Create an attachment from a file"""

    # File path
    path = os.path.join(os.path.dirname(__file__), '../data/edison.jpg')
    attachment = EmailAttachment.from_file(path)

    assert isinstance(attachment, EmailAttachment)
    assert attachment.filename == 'edison.jpg'
    assert attachment.mimetype == 'image/jpeg'

    # File object
    with open(path, 'rb') as f:
        attachment = EmailAttachment.from_file(f)

        assert isinstance(attachment, EmailAttachment)
        assert attachment.filename == 'edison.jpg'
        assert attachment.mimetype == 'image/jpeg'



"""
Mailer backend sends email via SMTP.
"""

from email import generator
from io import BytesIO
import smtplib
import socket
import ssl
import threading

from manhattan.mail import EmailTemplate
from manhattan.mail.backends.base import BaseMailer

__all__ = ['Mailer']


class Mailer(BaseMailer):
    """
    SMTP mailer
    """

    def __init__(
        self,
        host=None,
        port=None,
        username=None,
        password=None,
        use_tls=None,
        use_ssl=None,
        timeout=None,
        ssl_keyfile=None,
        ssl_certfile=None,
        *args,
        **kwargs
        ):
        super().__init__(*args, **kwargs)

        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.use_tls = use_tls
        self.use_ssl = use_ssl
        self.timeout = timeout
        self.ssl_keyfile = ssl_keyfile
        self.ssl_certfile = ssl_certfile

        assert not (self.use_ssl and self.use_tls), \
            '`use_tls` and `use_ssl` are mutually exclusive, \
only one can be true'

        self.connection = None
        self._lock = threading.RLock()

    def open(self):
        """Open a connection to the email server"""

        # Check if we already have a connection
        if self.connection:
            return

        # Build connection args
        connection_options = {'local_hostname': socket.getfqdn()}
        if self.timeout is not None:
            connection_options['timeout'] = self.timeout

        if self.use_ssl:
            connection_options['keyfile'] = self.ssl_keyfile
            connection_options['certfile'] = self.ssl_certfile

        try:
            # Attempt to connect to the email server
            connection_cls =  smtplib.SMTP
            if self.use_ssl:
                connection_cls = smtplib.SMTP_SSL

            self.connection = connection_cls(
                self.host,
                self.port,
                **connection_options
                )

            # Authentication
            if self.use_tls:

                # Check for a key file and certificate
                tls_args = {}
                if self.ssl_keyfile:
                    tls_args['keyfile'] = self.ssl_keyfile
                if self.ssl_certfile:
                    tls_args['certfile'] = self.ssl_certfile

                self.connection.starttls(**tls_args)

            if self.username and self.password:
                self.connection.login(self.username, self.password)

        except (smtplib.SMTPException, socket.error):
            self.connection = None

            if not self.fail_silently:
                raise

    def close(self):
        """Close the connection the email server"""

        # Check if we have a connection to close
        if self.connection is None:
            return

        # Attempt to close the connection
        try:
            try:
                self.connection.quit()
            except (ssl.SSLError, smtplib.SMTPServerDisconnected):
                self.connection.close()
            except smtplib.SMTPException:
                if not self.fail_silently:
                    raise
        finally:
            self.connection = None

    def send(self, messages):
        """Send all email messages via SMTP"""
        if isinstance(messages, EmailTemplate):
            messages = messages.merge()

        message_count = 0

        with self._lock:
            self.open()

            # Check we have a connection to send with
            if not self.connection:
                return

            for message in messages:
                try:

                    # Output the message using RFC valid line endings (CRLF)
                    fp = BytesIO()
                    g = generator.BytesGenerator(fp, mangle_from_=False)
                    g.flatten(message.message, linesep='\r\n')

                    # Attempt to send the message
                    self.connection.sendmail(
                        str(message.sender),
                        [str(r) for r in message.recipients],
                        fp.getvalue()
                    )
                    message_count += 1
                except smtplib.SMTPException:
                    if not self.fail_silently:
                        raise

            self.close()

        return message_count



import os
import pytest


from manhattan.mail import (
    EmailAttachment,
    EmailMessage
    )
from manhattan.mail.backends import smtp
try:
    from manhattan.mail.tests.smtp_settings import settings
except:
    settings = {}

# Fixtures

@pytest.fixture(scope='function')
def mailer():
    mailer = smtp.Mailer(**settings)
    return mailer


# Tests

def test_init():
    """Initialize an email contact"""
    mailer = smtp.Mailer(**settings)
    assert isinstance(mailer, smtp.Mailer)

def test_send_email_message(mailer):
    """Write an email message to memory (as opposed to actually sending them)"""
    if not settings:
        return

    path = os.path.join(os.path.dirname(__file__), '../data/edison.jpg')
    email_message = EmailMessage(
        ['ant@getme.co.uk'],
        'devs@getme.co.uk',
        'Hey',
        body_html='<html><body><p>Hello world</p></body></html>',
        attachments=[EmailAttachment.from_file(path)],
        headers={'Custom-header': 'custom-value'}
        )
    mailer.send([email_message])

    assert True



from manhattan.mail import EmailContact


def test_init():
    """Initialize an email contact"""
    contact = EmailContact('burt@example.com', 'Burt')
    assert isinstance(contact, EmailContact)

def test_eq():
    """Compare email contacts for equality"""
    contact_a = EmailContact('burt@example.com', 'Burt')
    contact_b = EmailContact('burt@example.com')
    contact_c = EmailContact('fred@example.com', 'Fred')

    assert contact_a == contact_b
    assert contact_a != contact_c

def test_hash():
    """Return a unique hash value for the contact"""
    contact_a = EmailContact('burt@example.com', 'Burt')
    contact_b = EmailContact('burt@example.com')
    assert hash(contact_a) == hash(contact_b)

def test_str():
    """Convert to an address string"""
    contact = EmailContact('burt@example.com', 'Burt')
    assert str(contact) == 'Burt <burt@example.com>'

def test_tuple():
    """Convert to a tuple"""
    contact = EmailContact('burt@example.com', 'Burt')
    assert tuple(contact) == ('burt@example.com', 'Burt', 'utf-8')

def test_normalize():
    """
    Convert a list of email address of the form string, tuple or `EmailContact`
    to a list of `EmailContact` instances.
    """
    contacts = EmailContact.normalize([
        'jojo@example.com',
        ('burt@example.com', 'Burt'),
        EmailContact('fred@example.com', 'Fred')
        ])

    assert len(contacts) == 3
    for contact in contacts:
        assert isinstance(contact, EmailContact)



from io import StringIO
import pytest

from manhattan.mail.backends import console


# Fixtures

@pytest.fixture(scope='function')
def mailer():
    stream = StringIO()
    mailer = console.Mailer(stream)
    mailer.stream = stream
    return mailer


# Utils

def write_message(message):
    stream = StringIO()
    msg_data = message.message.as_bytes()
    msg_data = msg_data.decode(message.encoding)
    stream.write('%s\n' % msg_data)
    stream.write('-' * 79)
    stream.write('\n')
    return stream.getvalue()

# Tests

def test_init():
    """Initialize an email contact"""
    mailer = console.Mailer()
    assert isinstance(mailer, console.Mailer)

def test_send_email_message(mailer, email_message_frozen):
    """Write an email message to memory (as opposed to actually sending them)"""
    mailer.send([email_message_frozen])
    assert mailer.stream.getvalue() == write_message(email_message_frozen)

def test_send_email_template(mailer, email_template_frozen):
    """Write an email template to memory (as opposed to actually sending them)"""
    email_messages = email_template_frozen.merge()
    mailer.send(email_template_frozen)

    output = write_message(email_messages[0])
    output += write_message(email_messages[1])
    mailer.stream.getvalue() == output



"""
Mailer backend that writes email messages to the console instead of sending
them.
"""

import sys
import threading

from manhattan.mail import EmailTemplate
from manhattan.mail.backends.base import BaseMailer

__all__ = ['Mailer']


class Mailer(BaseMailer):
    """
    Mailer that writes emails messages to the console.
    """

    def __init__(self, stream=None, *args, **kwargs):
        self.stream = stream or sys.stdout
        self._lock = threading.RLock()
        super().__init__(*args, **kwargs)

    def send(self, messages):
        """Write all email messages to the stream"""
        if isinstance(messages, EmailTemplate):
            messages = messages.merge()

        message_count = 0

        with self._lock:
            try:
                for message in messages:
                    self._write_message(message)
                    self.stream.flush()
                    message_count += 1
            except Exception:
                if not self.fail_silently:
                    raise

        return message_count

    def _write_message(self, message):
        """Write a message to the stream"""
        msg_data = message.message.as_bytes()
        msg_data = msg_data.decode(message.encoding)
        self.stream.write('%s\n' % msg_data)
        self.stream.write('-' * 79)
        self.stream.write('\n')



import os

import pytest

from manhattan.mail import (
    EmailAttachment,
    EmailContact,
    EmailMessage,
    EmailTemplate
    )

__all__ = [
    'email_message',
    'email_message_frozen',
    'email_template',
    'email_template_frozen'
    ]


# Classes

class FrozenEmailMessage(EmailMessage):

    def __init__(self, email_message):
        self._message = email_message.message
        self.encoding = email_message.encoding

    @property
    def message(self):
        return self._message


class FrozenEmailTemplate(EmailTemplate):

    def __init__(self, email_template):
        self.email_messages = [
            FrozenEmailMessage(e) for e in email_template.merge()]

    def merge(self):
        return self.email_messages


# Fixtures

@pytest.fixture(scope='function')
def email_message():
    """Return an email message"""
    path = os.path.join(os.path.dirname(__file__), 'data/edison.jpg')
    message = EmailMessage(
        ['burt@example.com'],
        'fred@example.com',
        'Hey',
        body_html='<html><body><p>Hello world</p></body></html>',
        cc=['jojo@example.com'],
        attachments=[EmailAttachment.from_file(path)],
        headers={'Custom-header': 'custom-value'}
        )
    return message

@pytest.fixture(scope='function')
def email_message_frozen(email_message):
    """Return a frozen email message"""
    return FrozenEmailMessage(email_message)

@pytest.fixture(scope='function')
def email_template():
    """Return an email template"""
    path = os.path.join(os.path.dirname(__file__), 'data/edison.jpg')

    template_path = os.path.join(os.path.dirname(__file__), 'data')
    template_map = {}
    for template_name in ['example', 'example_include']:
        with open(os.path.join(template_path, template_name + '.html')) as f:
            template_map[template_name] = f.read()

    template = EmailTemplate(
        ['burt@example.com', 'appa@example.com'],
        'fred@example.com',
        'Hey {{ name }}',
        'example',
        template_map=template_map,
        css='h1 { color: #f00; }',
        recipient_vars={
            'burt@example.com': {'name': 'Burt'},
            'appa@example.com': {'name': 'Appa'},
        },
        global_vars={'company': 'Getme'},
        cc=['jojo@example.com'],
        bcc=['shhh@example.com'],
        attachments=[EmailAttachment.from_file(path)],
        headers={'Custom-header': 'custom-value'}
        )

    return template

@pytest.fixture(scope='function')
def email_template_frozen(email_template):
    """Return an frozen email template"""
    return FrozenEmailTemplate(email_template)



from io import StringIO
import os
import pytest

from manhattan.mail.backends import file


# Fixtures

@pytest.fixture(scope='function')
def mailer():
    mailer = file.Mailer(os.path.join(os.getcwd(), 'test-output/test.eml'))
    return mailer


# Utils

def write_message(message):
    stream = StringIO()
    msg_data = message.message.as_bytes()
    msg_data = msg_data.decode(message.encoding)
    stream.write('%s\n' % msg_data)
    stream.write('-' * 79)
    stream.write('\n')
    return stream.getvalue()

# Tests

def test_init():
    """Initialize an email contact"""
    mailer = file.Mailer('some_file_path.eml')
    assert isinstance(mailer, file.Mailer)

def test_send_email_message(mailer, email_message_frozen):
    """Write an email message to file (as opposed to actually sending them)"""
    mailer.send([email_message_frozen])
    with open(mailer.file_path) as f:
        assert f.read() == write_message(email_message_frozen)

def test_send_email_template(mailer, email_template_frozen):
    """Write an email template to file (as opposed to actually sending them)"""
    email_messages = email_template_frozen.merge()
    mailer.send(email_template_frozen)

    output = write_message(email_messages[0])
    output = write_message(email_messages[1])

    with open(mailer.file_path) as f:
        assert f.read() == output


from email.mime.image import MIMEImage
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
import os

from manhattan.mail import EmailAttachment, EmailContact, EmailMessage


def test_init():
    """Initialize an email message"""
    message = EmailMessage(
        ['burt@example.com'],
        'fred@example.com',
        'Hey',
        body_text='Hello world'
        )
    assert isinstance(message, EmailMessage)

def test_auto_text():
    """Generate a text alternative if no text body is provided"""
    message = EmailMessage(
        'burt@example.com',
        'fred@example.com',
        'Hey',
        body_html='<html><body><p>Hello world</p></body></html>'
        )
    assert message.body_text == 'Hello world\n\n'

def test_attach():
    """Attach a file to the email message"""
    message = EmailMessage(
        ['burt@example.com'],
        'fred@example.com',
        'Hey',
        body_text='Hello world'
        )
    path = os.path.join(os.path.dirname(__file__), '../data/edison.jpg')
    attachment = message.attach(path)

    assert len(message.attachments) == 1
    assert message.attachments[-1] == attachment

def test_message():
    """Return an email message"""
    path = os.path.join(os.path.dirname(__file__), '../data/edison.jpg')
    message = EmailMessage(
        ['burt@example.com'],
        'fred@example.com',
        'Hey',
        body_html='<html><body><p>Hello world</p></body></html>',
        cc=['jojo@example.com'],
        attachments=[EmailAttachment.from_file(path)],
        headers={'Custom-header': 'custom-value'}
        )
    msg = message.message
    payload =  msg.get_payload()

    assert isinstance(msg, MIMEMultipart)
    assert msg['Subject'] == message.subject
    assert msg['From'] == str(message.sender)
    assert msg['To'] == ', '.join([str(t) for t in message.to])
    assert msg['Cc'] == ', '.join([str(c) for c in message.cc])

    assert len(payload) == 2
    assert payload[0].get_payload()[0].get_payload() == 'Hello world\n\n'
    assert payload[0].get_payload()[1].get_payload() == \
            '<html><body><p>Hello world</p></body></html>'
    assert isinstance(payload[1], MIMEImage)

def test_recipients():
    """Return a list of all recipients for an email message (to, cc, bcc)"""
    message = EmailMessage(
        ['burt@example.com'],
        'fred@example.com',
        'Hey',
        body_text='Hello world',
        cc=['jojo@example.com'],
        bcc=['appa@example.com']
        )

    assert message.recipients == EmailContact.normalize([
        'burt@example.com',
        'jojo@example.com',
        'appa@example.com'
        ])



from email.mime.image import MIMEImage
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
import os

from jinja2 import Environment, FileSystemLoader

from manhattan.mail import (
    EmailAttachment,
    EmailContact,
    EmailMessage,
    EmailTemplate
    )



class CustomEmailTemplate(EmailTemplate):
    """
    An `EmailTemplate` class set up for the test environment.
    """

    def get_jinja_env(self):
        path = os.path.join(os.path.dirname(__file__), '../data/')
        return Environment(
            loader=FileSystemLoader(path),
            autoescape=True
            )


# Tests

def test_init():
    """Initialize an email template"""
    template = EmailTemplate(
        ['burt@example.com'],
        'fred@example.com',
        'Hey {{ name }}',
        'example.html'
        )
    assert isinstance(template, EmailTemplate)

def test_attach():
    """Attach a file to the email message"""
    message = EmailTemplate(
        ['burt@example.com'],
        'fred@example.com',
        'Hey {{ name }}',
        'example.html'
        )
    path = os.path.join(os.path.dirname(__file__), '../data/edison.jpg')
    attachment = message.attach(path)

    assert len(message.attachments) == 1
    assert message.attachments[-1] == attachment

def test_merge_template_file():
    """Mail merge using a template file to create list of email messages"""
    path = os.path.join(os.path.dirname(__file__), '../data/edison.jpg')
    template = CustomEmailTemplate(
        ['burt@example.com', 'appa@example.com'],
        'fred@example.com',
        'Hey {{ name }}',
        'example.html',
        css='h1 { color: #f00; }',
        recipient_vars={
            'burt@example.com': {'name': 'Burt'},
            'appa@example.com': {'name': 'Appa'},
        },
        global_vars={'company': 'Getme'},
        cc=['jojo@example.com'],
        bcc=['shhh@example.com'],
        attachments=[EmailAttachment.from_file(path)],
        headers={'Custom-header': 'custom-value'}
        )
    messages = template.merge()

    assert len(messages) == 2
    assert messages[0].body_html == '''<html>
    <head><style type="text/css">h1 { color: #f00; }</style></head>
<body>
        <h1 style="color:#f00">Getme</h1>
        <p>Burt</p>
        <p>Footer</p>
    </body>
</html>
'''
    assert len(messages[0].attachments) == 1

    assert messages[1].body_html == '''<html>
    <head><style type="text/css">h1 { color: #f00; }</style></head>
<body>
        <h1 style="color:#f00">Getme</h1>
        <p>Appa</p>
        <p>Footer</p>
    </body>
</html>
'''
    assert len(messages[1].attachments) == 1

def test_merge_template_map():
    """Mail merge using a template file to create list of email messages"""
    path = os.path.join(os.path.dirname(__file__), '../data/edison.jpg')

    template_path = os.path.join(os.path.dirname(__file__), '../data')
    template_map = {}
    for template_name in ['example', 'example_include']:
        with open(os.path.join(template_path, template_name + '.html')) as f:
            template_map[template_name] = f.read()

    template = EmailTemplate(
        ['burt@example.com', 'appa@example.com'],
        'fred@example.com',
        'Hey {{ name }}',
        'example',
        template_map=template_map,
        css='h1 { color: #f00; }',
        recipient_vars={
            'burt@example.com': {'name': 'Burt'},
            'appa@example.com': {'name': 'Appa'},
        },
        global_vars={'company': 'Getme'},
        cc=['jojo@example.com'],
        bcc=['shhh@example.com'],
        attachments=[EmailAttachment.from_file(path)],
        headers={'Custom-header': 'custom-value'}
        )
    messages = template.merge()

    assert len(messages) == 2
    assert messages[0].body_html == '''<html>
    <head><style type="text/css">h1 { color: #f00; }</style></head>
<body>
        <h1 style="color:#f00">Getme</h1>
        <p>Burt</p>
        <p>Footer</p>
    </body>
</html>
'''
    assert len(messages[0].attachments) == 1

    assert messages[1].body_html == '''<html>
    <head><style type="text/css">h1 { color: #f00; }</style></head>
<body>
        <h1 style="color:#f00">Getme</h1>
        <p>Appa</p>
        <p>Footer</p>
    </body>
</html>
'''
    assert len(messages[1].attachments) == 1
